"""Builder superclass for all builders."""

from __future__ import annotations

import pickle
import re
import time
from contextlib import nullcontext
from pathlib import Path
from typing import TYPE_CHECKING, final

from docutils import nodes
from docutils.utils import DependencyList

from sphinx._cli.util.colour import bold
from sphinx.deprecation import _deprecation_warning
from sphinx.environment import (
    CONFIG_CHANGED_REASON,
    CONFIG_OK,
    _CurrentDocument,
)
from sphinx.environment.adapters.asset import ImageAdapter
from sphinx.errors import SphinxError
from sphinx.locale import __
from sphinx.util import get_filetype, logging
from sphinx.util._importer import import_object
from sphinx.util._pathlib import _StrPathProperty
from sphinx.util.build_phase import BuildPhase
from sphinx.util.display import progress_message, status_iterator
from sphinx.util.docutils import _parse_str_to_doctree
from sphinx.util.i18n import CatalogRepository, docname_to_domain
from sphinx.util.osutil import ensuredir, relative_uri, relpath
from sphinx.util.parallel import (
    ParallelTasks,
    SerialTasks,
    make_chunks,
    parallel_available,
)

# side effect: registers roles and directives
from sphinx import directives  # NoQA: F401  isort:skip
from sphinx import roles  # NoQA: F401  isort:skip

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence, Set
    from gettext import NullTranslations
    from typing import Any, ClassVar, Literal

    from docutils.nodes import Node

    from sphinx.application import Sphinx
    from sphinx.config import Config
    from sphinx.environment import (
        BuildEnvironment,
    )
    from sphinx.events import EventManager
    from sphinx.util.i18n import CatalogInfo
    from sphinx.util.tags import Tags


logger = logging.getLogger(__name__)


class Builder:
    """Builds target formats from the reST sources."""

    #: The builder's name.
    #: This is the value used to select builders on the command line.
    name: ClassVar[str] = ''
    #: The builder's output format, or '' if no document output is produced.
    #: This is commonly the file extension, e.g. "html",
    #: though any string value is accepted.
    #: The builder's format string can be used by various components
    #: such as :class:`.SphinxPostTransform` or extensions to determine
    #: their compatibility with the builder.
    format: ClassVar[str] = ''
    #: The message emitted upon successful build completion.
    #: This can be a printf-style template string
    #: with the following keys: ``outdir``, ``project``
    epilog: ClassVar[str] = ''

    #: default translator class for the builder.  This can be overridden by
    #: :py:meth:`~sphinx.application.Sphinx.set_translator`.
    default_translator_class: ClassVar[type[nodes.NodeVisitor]]
    # doctree versioning method
    versioning_method: ClassVar[str] = 'none'
    versioning_compare: ClassVar[bool] = False
    #: Whether it is safe to make parallel :meth:`~.Builder.write_doc` calls.
    allow_parallel: ClassVar[bool] = False
    # support translation
    use_message_catalog: ClassVar[bool] = True

    #: The list of MIME types of image formats supported by the builder.
    #: Image files are searched in the order in which they appear here.
    supported_image_types: ClassVar[list[str]] = []
    #: The builder can produce output documents that may fetch external images when opened.
    supported_remote_images: ClassVar[bool] = False
    #: The file format produced by the builder allows images to be embedded using data-URIs.
    supported_data_uri_images: ClassVar[bool] = False

    phase: BuildPhase = BuildPhase.INITIALIZATION

    srcdir = _StrPathProperty()
    confdir = _StrPathProperty()
    outdir = _StrPathProperty()
    doctreedir = _StrPathProperty()

    def __init__(self, app: Sphinx, env: BuildEnvironment) -> None:
        self.srcdir = app.srcdir
        self.confdir = app.confdir
        self.outdir = app.outdir
        self.doctreedir = app.doctreedir
        ensuredir(self.doctreedir)

        self._app: Sphinx = app
        self.env: BuildEnvironment = env
        self.env.set_versioning_method(self.versioning_method, self.versioning_compare)
        self.events: EventManager = app.events
        self.config: Config = app.config
        self.tags: Tags = app.tags
        self.tags.add(self.format)
        self.tags.add(self.name)
        self.tags.add(f'format_{self.format}')
        self.tags.add(f'builder_{self.name}')
        self._registry = app.registry

        # images that need to be copied over (source -> dest)
        self.images: dict[str, str] = {}
        # basename of images directory
        self.imagedir = ''
        # relative path to image directory from current docname (used at writing docs)
        self.imgpath = ''

        # these get set later
        self.parallel_ok = False
        self.finish_tasks: Any = None

    @property
    def app(self) -> Sphinx:
        cls_module = self.__class__.__module__
        cls_name = self.__class__.__qualname__
        _deprecation_warning(cls_module, f'{cls_name}.app', remove=(11, 0))
        return self._app

    @property
    def _translator(self) -> NullTranslations | None:
        return self._app.translator

    def get_translator_class(self, *args: Any) -> type[nodes.NodeVisitor]:
        """Return a class of translator."""
        return self._registry.get_translator_class(self)

    def create_translator(self, *args: Any) -> nodes.NodeVisitor:
        """Return an instance of translator.

        This method returns an instance of ``default_translator_class`` by default.
        Users can replace the translator class with ``app.set_translator()`` API.
        """
        return self._registry.create_translator(self, *args)

    # helper methods
    def init(self) -> None:
        """Load necessary templates and perform initialization.  The default
        implementation does nothing.
        """
        pass

    def create_template_bridge(self) -> None:
        """Return the template bridge configured."""
        if self.config.template_bridge:
            template_bridge_cls = import_object(
                self.config.template_bridge,
                source='template_bridge setting',
            )
            self.templates = template_bridge_cls()
        else:
            from sphinx.jinja2glue import BuiltinTemplateLoader

            self.templates = BuiltinTemplateLoader()

    def get_target_uri(self, docname: str, typ: str | None = None) -> str:
        """Return the target URI for a document name.

        *typ* can be used to qualify the link characteristic for individual
        builders.
        """
        raise NotImplementedError

    def get_relative_uri(self, from_: str, to: str, typ: str | None = None) -> str:
        """Return a relative URI between two source filenames.

        :raises: :exc:`!NoUri` if there's no way to return a sensible URI.
        """
        return relative_uri(
            self.get_target_uri(from_),
            self.get_target_uri(to, typ),
        )

    def get_outdated_docs(self) -> str | Iterable[str]:
        """Return an iterable of output files that are outdated, or a string
        describing what an update build will build.

        If the builder does not output individual files corresponding to
        source files, return a string here.  If it does, return an iterable
        of those files that need to be written.
        """
        raise NotImplementedError

    def get_asset_paths(self) -> list[str]:
        """Return list of paths for assets (ex. templates, CSS, etc.)."""
        return []

    def post_process_images(self, doctree: Node) -> None:
        """Pick the best candidate for all image URIs."""
        images = ImageAdapter(self.env)
        for node in doctree.findall(nodes.image):
            if '?' in node['candidates']:
                # don't rewrite nonlocal image URIs
                continue
            if '*' not in node['candidates']:
                for imgtype in self.supported_image_types:
                    candidate = node['candidates'].get(imgtype, None)
                    if candidate:
                        break
                else:
                    mimetypes = sorted(node['candidates'])
                    image_uri = images.get_original_image_uri(node['uri'])
                    if mimetypes:
                        logger.warning(
                            __('a suitable image for %s builder not found: %s (%s)'),
                            self.name,
                            mimetypes,
                            image_uri,
                            location=node,
                        )
                    else:
                        logger.warning(
                            __('a suitable image for %s builder not found: %s'),
                            self.name,
                            image_uri,
                            location=node,
                        )
                    continue
                node['uri'] = candidate
            else:
                candidate = node['uri']
            if candidate not in self.env.images:
                # non-existing URI; let it alone
                continue
            self.images[candidate] = self.env.images[candidate][1]

    # compile po methods

    def compile_catalogs(self, catalogs: set[CatalogInfo], message: str) -> None:
        if not self.config.gettext_auto_build:
            return

        def cat2relpath(cat: CatalogInfo, srcdir: Path = self.srcdir) -> str:
            return Path(relpath(cat.mo_path, srcdir)).as_posix()

        logger.info(bold(__('building [mo]: ')) + message)  # NoQA: G003
        for catalog in status_iterator(
            catalogs,
            __('writing output... '),
            'darkgreen',
            len(catalogs),
            self.config.verbosity,
            stringify_func=cat2relpath,
        ):
            catalog.write_mo(
                self.config.language, self.config.gettext_allow_fuzzy_translations
            )

    def compile_all_catalogs(self) -> None:
        repo = CatalogRepository(
            self.srcdir,
            self.config.locale_dirs,
            self.config.language,
            self.config.source_encoding,
        )
        message = __('all of %d po files') % len(list(repo.catalogs))
        self.compile_catalogs(set(repo.catalogs), message)

    def compile_specific_catalogs(self, specified_files: Iterable[Path]) -> None:
        env = self.env
        gettext_compact = self.config.gettext_compact

        domains = {
            docname_to_domain(docname, gettext_compact) if docname else None
            for file in specified_files
            if (docname := env.path2doc(file))
        }
        catalogs = set()
        repo = CatalogRepository(
            self.srcdir,
            self.config.locale_dirs,
            self.config.language,
            self.config.source_encoding,
        )
        for catalog in repo.catalogs:
            if catalog.domain in domains and catalog.is_outdated():
                catalogs.add(catalog)
        message = __('targets for %d po files that are specified') % len(catalogs)
        self.compile_catalogs(catalogs, message)

    # TODO(stephenfin): This would make more sense as 'compile_outdated_catalogs'
    def compile_update_catalogs(self) -> None:
        repo = CatalogRepository(
            self.srcdir,
            self.config.locale_dirs,
            self.config.language,
            self.config.source_encoding,
        )
        catalogs = {c for c in repo.catalogs if c.is_outdated()}
        message = __('targets for %d po files that are out of date') % len(catalogs)
        self.compile_catalogs(catalogs, message)

    # build methods

    @final
    def build_all(self) -> None:
        """Build all source files."""
        self.compile_all_catalogs()

        self.build(None, summary=__('all source files'), method='all')

    @final
    def build_specific(self, filenames: Sequence[Path]) -> None:
        """Only rebuild as much as needed for changes in the *filenames*."""
        docnames: list[str] = []

        filenames = [Path(filename).resolve() for filename in filenames]
        for filename in filenames:
            if not filename.is_file():
                logger.warning(
                    __('file %r given on command line does not exist, '), filename
                )
                continue

            if not filename.is_relative_to(self.srcdir):
                logger.warning(
                    __(
                        'file %r given on command line is not under the '
                        'source directory, ignoring'
                    ),
                    filename,
                )
                continue

            docname = self.env.path2doc(filename)
            if not docname:
                logger.warning(
                    __(
                        'file %r given on command line is not a valid '
                        'document, ignoring'
                    ),
                    filename,
                )
                continue

            docnames.append(docname)

        self.compile_specific_catalogs(filenames)

        self.build(
            docnames,
            summary=__('%d source files given on command line') % len(docnames),
            method='specific',
        )

    @final
    def build_update(self) -> None:
        """Only rebuild what was changed or added since last build."""
        self.compile_update_catalogs()

        to_build = self.get_outdated_docs()
        if isinstance(to_build, str):
            self.build(['__all__'], summary=to_build, method='update')
        else:
            to_build = set(to_build)
            self.build(
                to_build,
                summary=__('targets for %d source files that are out of date')
                % len(to_build),
                method='update',
            )

    @final
    def build(
        self,
        docnames: Iterable[str] | None,
        summary: str | None = None,
        method: Literal['all', 'specific', 'update'] = 'update',
    ) -> None:
        """Main build method, usually called by a specific ``build_*`` method.

        First updates the environment, and then calls
        :meth:`!write`.
        """
        if summary:
            logger.info(bold(__('building [%s]: ')) + summary, self.name)  # NoQA: G003

        # while reading, collect all warnings from docutils
        with (
            nullcontext()
            if self._app._exception_on_warning
            else logging.pending_warnings()
        ):
            updated_docnames = set(self.read())

        doccount = len(updated_docnames)
        logger.info(bold(__('looking for now-outdated files... ')), nonl=True)
        updated_docnames.update(self.env.check_dependents(self._app, updated_docnames))
        outdated = len(updated_docnames) - doccount
        if outdated:
            logger.info(__('%d found'), outdated)
        else:
            logger.info(__('none found'))

        if updated_docnames:
            # save the environment
            from sphinx.application import ENV_PICKLE_FILENAME

            with (
                progress_message(__('pickling environment')),
                open(self.doctreedir / ENV_PICKLE_FILENAME, 'wb') as f,
            ):
                pickle.dump(self.env, f, pickle.HIGHEST_PROTOCOL)

            # global actions
            self.phase = BuildPhase.CONSISTENCY_CHECK
            with progress_message(__('checking consistency')):
                self.env.check_consistency()
        else:
            if method == 'update' and not docnames:
                logger.info(bold(__('no targets are out of date.')))

        self.phase = BuildPhase.RESOLVING

        # filter "docnames" (list of outdated files) by the updated
        # found_docs of the environment; this will remove docs that
        # have since been removed
        if docnames and docnames != ['__all__']:
            docnames = set(docnames) & self.env.found_docs

        # determine if we can write in parallel
        if parallel_available and self._app.parallel > 1 and self.allow_parallel:
            self.parallel_ok = self._app.is_parallel_allowed('write')
        else:
            self.parallel_ok = False

        #  create a task executor to use for misc. "finish-up" tasks
        # if self.parallel_ok:
        #     self.finish_tasks = ParallelTasks(self._app.parallel)
        # else:
        # for now, just execute them serially
        self.finish_tasks = SerialTasks()

        # write all "normal" documents (or everything for some builders)
        self.write(docnames, updated_docnames, method)

        # finish (write static files etc.)
        self.finish()

        # wait for all tasks
        self.finish_tasks.join()

    @final
    def read(self) -> list[str]:
        """(Re-)read all files new or changed since last update.

        Store all environment docnames in the canonical format (ie using SEP as
        a separator in place of os.path.sep).
        """
        logger.info(bold(__('updating environment: ')), nonl=True)

        self.env.find_files(self.config, self)
        updated = self.env.config_status != CONFIG_OK
        added, changed, removed = self.env.get_outdated_files(updated)

        # allow user intervention as well
        for docs in self.events.emit(
            'env-get-outdated', self.env, added, changed, removed
        ):
            changed.update(set(docs) & self.env.found_docs)

        # if files were added or removed, all documents with globbed toctrees
        # must be reread
        if added or removed:
            # ... but not those that already were removed
            changed.update(self.env.glob_toctrees & self.env.found_docs)

        if updated:  # explain the change iff build config status was not ok
            reason = CONFIG_CHANGED_REASON.get(self.env.config_status, '') + (
                self.env.config_status_extra or ''
            )
            logger.info('[%s] ', reason, nonl=True)

        logger.info(
            __('%s added, %s changed, %s removed'),
            len(added),
            len(changed),
            len(removed),
        )

        # clear all files no longer present
        for docname in removed:
            self.events.emit('env-purge-doc', self.env, docname)
            self.env.clear_doc(docname)

        # read all new and changed files
        docnames = sorted(added | changed)
        # allow changing and reordering the list of docs to read
        self.events.emit('env-before-read-docs', self.env, docnames)

        # check if we should do parallel or serial read
        if parallel_available and self._app.parallel > 1:
            par_ok = self._app.is_parallel_allowed('read')
        else:
            par_ok = False

        if par_ok:
            self._read_parallel(docnames, nproc=self._app.parallel)
        else:
            self._read_serial(docnames)

        if self.config.master_doc not in self.env.all_docs:
            from sphinx.project import EXCLUDE_PATHS
            from sphinx.util.matching import _translate_pattern

            master_doc_path = self.env.doc2path(self.config.master_doc)
            master_doc_canon = master_doc_path.as_posix()
            for pat in EXCLUDE_PATHS:
                if not re.match(_translate_pattern(pat), master_doc_canon):
                    continue
                msg = __(
                    'Sphinx is unable to load the master document (%s) '
                    'because it matches a built-in exclude pattern %r. '
                    'Please move your master document to a different location.'
                )
                raise SphinxError(msg % (master_doc_path, pat))
            for pat in self.config.exclude_patterns:
                if not re.match(_translate_pattern(pat), master_doc_canon):
                    continue
                msg = __(
                    'Sphinx is unable to load the master document (%s) '
                    'because it matches an exclude pattern specified '
                    'in conf.py, %r. '
                    'Please remove this pattern from conf.py.'
                )
                raise SphinxError(msg % (master_doc_path, pat))
            if set(self.config.include_patterns) != {'**'} and not any(
                re.match(_translate_pattern(pat), master_doc_canon)
                for pat in self.config.include_patterns
            ):
                msg = __(
                    'Sphinx is unable to load the master document (%s) '
                    'because it is not included in the custom include_patterns = %r. '
                    'Ensure that a pattern in include_patterns matches the '
                    'master document.'
                )
                raise SphinxError(msg % (master_doc_path, self.config.include_patterns))
            msg = __(
                'Sphinx is unable to load the master document (%s). '
                'The master document must be within the source directory '
                'or a subdirectory of it.'
            )
            raise SphinxError(msg % master_doc_path)

        for retval in self.events.emit('env-updated', self.env):
            if retval is not None:
                docnames.extend(retval)

        # workaround: marked as okay to call builder.read() twice in same process
        self.env.config_status = CONFIG_OK

        return sorted(docnames)

    def _read_serial(self, docnames: list[str]) -> None:
        for docname in status_iterator(
            docnames,
            __('reading sources... '),
            'purple',
            len(docnames),
            self.config.verbosity,
        ):
            # remove all inventory entries for that file
            self.events.emit('env-purge-doc', self.env, docname)
            self.env.clear_doc(docname)
            self.read_doc(docname)

    def _read_parallel(self, docnames: list[str], nproc: int) -> None:
        chunks = make_chunks(docnames, nproc)

        # create a status_iterator to step progressbar after reading a document
        # (see: ``merge()`` function)
        progress = status_iterator(
            chunks,
            __('reading sources... '),
            'purple',
            len(chunks),
            self.config.verbosity,
        )

        # clear all outdated docs at once
        for docname in docnames:
            self.events.emit('env-purge-doc', self.env, docname)
            self.env.clear_doc(docname)

        def read_process(docs: list[str]) -> bytes:
            self.env._app = self._app
            for docname in docs:
                self.read_doc(docname, _cache=False)
            # allow pickling self to send it back
            return pickle.dumps(self.env, pickle.HIGHEST_PROTOCOL)

        def merge(docs: list[str], otherenv: bytes) -> None:
            env = pickle.loads(otherenv)
            self.env.merge_info_from(docs, env, self._app)

            next(progress)

        tasks = ParallelTasks(nproc)
        for chunk in chunks:
            tasks.add_task(read_process, chunk, merge)

        # make sure all threads have finished
        tasks.join()
        logger.info('')

    @final
    def read_doc(self, docname: str, *, _cache: bool = True) -> None:
        """Parse a file and add/update inventory entries for the doctree."""
        env = self.env
        env.prepare_settings(docname)

        # Add confdir/docutils.conf to dependencies list if exists
        docutils_conf = self.confdir / 'docutils.conf'
        if docutils_conf.is_file():
            env.note_dependency(docutils_conf)

        filename = env.doc2path(docname)

        # read the source file
        content = filename.read_text(encoding=env.settings['input_encoding'])

        # TODO: move the "source-read" event to here.

        filetype = get_filetype(self.config.source_suffix, filename)
        parser = self._registry.create_source_parser(
            filetype, config=self.config, env=env
        )
        doctree = _parse_str_to_doctree(
            content,
            filename=filename,
            default_role=self.config.default_role,
            default_settings=env.settings,
            env=env,
            events=self.events,
            parser=parser,
            transforms=self._registry.get_transforms(),
        )

        # store time of reading, for outdated files detection
        env.all_docs[docname] = time.time_ns() // 1_000

        # cleanup
        env.current_document = _CurrentDocument()
        env.ref_context.clear()

        self.write_doctree(docname, doctree, _cache=_cache)

    @final
    def write_doctree(
        self,
        docname: str,
        doctree: nodes.document,
        *,
        _cache: bool = True,
    ) -> None:
        """Write the doctree to a file, to be used as a cache by re-builds."""
        # make it pickleable
        doctree.reporter = None  # type: ignore[assignment]
        doctree.transformer = None  # type: ignore[assignment]

        # Create a copy of settings object before modification because it is
        # shared with other documents.
        doctree.settings = doctree.settings.copy()
        doctree.settings.warning_stream = None
        doctree.settings.env = None
        doctree.settings.record_dependencies = DependencyList()

        doctree_filename = self.doctreedir / f'{docname}.doctree'
        doctree_filename.parent.mkdir(parents=True, exist_ok=True)
        with open(doctree_filename, 'wb') as f:
            pickle.dump(doctree, f, pickle.HIGHEST_PROTOCOL)

        # When Sphinx is running in parallel mode, ``write_doctree()`` is invoked
        # in the context of a process worker, and thus it does not make sense to
        # pickle the doctree and send it to the main process
        if _cache:
            self.env._write_doc_doctree_cache[docname] = doctree

    @final
    def write(
        self,
        build_docnames: Iterable[str] | None,
        updated_docnames: Iterable[str],
        method: Literal['all', 'specific', 'update'] = 'update',
    ) -> None:
        """Write builder specific output files."""
        env = self.env

        # Allow any extensions to perform setup for writing
        self.events.emit('write-started', self)

        if build_docnames is None or build_docnames == ['__all__']:
            # build_all
            build_docnames = env.found_docs
        if method == 'update':
            # build updated ones as well
            docnames = set(build_docnames) | set(updated_docnames)
        else:
            docnames = set(build_docnames)
        if docnames:
            logger.debug(__('docnames to write: %s'), ', '.join(sorted(docnames)))
        else:
            logger.debug(__('no docnames to write!'))

        # add all toctree-containing files that may have changed
        docnames |= {
            toc_docname
            for docname in docnames
            for toc_docname in env.files_to_rebuild.get(docname, ())
            if toc_docname in env.found_docs
        }

        # sort to ensure deterministic toctree generation
        env.toctree_includes = dict(sorted(env.toctree_includes.items()))

        with progress_message(__('preparing documents')):
            self.prepare_writing(docnames)

        with progress_message(__('copying assets'), nonl=False):
            self.copy_assets()

        if docnames:
            self.write_documents(docnames)

    def write_documents(self, docnames: Set[str]) -> None:
        """Write all documents in *docnames*.

        This method can be overridden if a builder does not create
        output files for each document.
        """
        sorted_docnames = sorted(docnames)
        if self.parallel_ok:
            # number of subprocesses is parallel-1 because the main process
            # is busy loading doctrees and doing write_doc_serialized()
            self._write_parallel(sorted_docnames, nproc=self._app.parallel - 1)
        else:
            self._write_serial(sorted_docnames)

    def _write_serial(self, docnames: Sequence[str]) -> None:
        with (
            nullcontext()
            if self._app._exception_on_warning
            else logging.pending_warnings()
        ):
            for docname in status_iterator(
                docnames,
                __('writing output... '),
                'darkgreen',
                len(docnames),
                self.config.verbosity,
            ):
                _write_docname(docname, env=self.env, builder=self, tags=self.tags)

    def _write_parallel(self, docnames: Sequence[str], nproc: int) -> None:
        def write_process(docs: list[tuple[str, nodes.document]]) -> None:
            self.phase = BuildPhase.WRITING
            for docname, doctree in docs:
                self.write_doc(docname, doctree)

        # warm up caches/compile templates using the first document
        firstname, docnames = docnames[0], docnames[1:]
        _write_docname(firstname, env=self.env, builder=self, tags=self.tags)

        tasks = ParallelTasks(nproc)
        chunks = make_chunks(docnames, nproc)

        # create a status_iterator to step progressbar after writing a document
        # (see: ``on_chunk_done()`` function)
        progress = status_iterator(
            chunks,
            __('writing output... '),
            'darkgreen',
            len(chunks),
            self.config.verbosity,
        )

        def on_chunk_done(args: list[tuple[str, nodes.document]], result: None) -> None:
            next(progress)

        self.phase = BuildPhase.RESOLVING
        for chunk in chunks:
            arg = []
            for docname in chunk:
                doctree = self.env.get_and_resolve_doctree(
                    docname, self, tags=self.tags
                )
                self.write_doc_serialized(docname, doctree)
                arg.append((docname, doctree))
            tasks.add_task(write_process, arg, on_chunk_done)

        # make sure all threads have finished
        tasks.join()
        logger.info('')

    def prepare_writing(self, docnames: Set[str]) -> None:
        """A place where you can add logic before :meth:`write_doc` is run"""
        pass

    def copy_assets(self) -> None:
        """Where assets (images, static files, etc) are copied before writing"""
        pass

    def write_doc(self, docname: str, doctree: nodes.document) -> None:
        """Write the output file for the document

        :param docname: the :term:`docname <document name>`.
        :param doctree: defines the content to be written.

        The output filename must be determined within this method,
        typically by calling :meth:`~.Builder.get_target_uri`
        or :meth:`~.Builder.get_relative_uri`.
        """
        raise NotImplementedError

    def write_doc_serialized(self, docname: str, doctree: nodes.document) -> None:
        """Handle parts of write_doc that must be called in the main process
        if parallel build is active.
        """
        pass

    def finish(self) -> None:
        """Finish the building process.

        The default implementation does nothing.
        """
        pass

    def cleanup(self) -> None:
        """Cleanup any resources.

        The default implementation does nothing.
        """
        pass

    def get_builder_config(self, option: str, default: str) -> Any:
        """Return a builder specific option.

        This method allows customization of common builder settings by
        inserting the name of the current builder in the option key.
        If the key does not exist, use default as builder name.
        """
        # At the moment, only XXX_use_index is looked up this way.
        # Every new builder variant must be registered in Config.config_values.
        try:
            optname = f'{self.name}_{option}'
            return getattr(self.config, optname)
        except AttributeError:
            optname = f'{default}_{option}'
            return getattr(self.config, optname)


def _write_docname(
    docname: str,
    /,
    *,
    env: BuildEnvironment,
    builder: Builder,
    tags: Tags,
) -> None:
    """Write a single document."""
    builder.phase = BuildPhase.RESOLVING
    doctree = env.get_and_resolve_doctree(docname, builder=builder, tags=tags)
    builder.phase = BuildPhase.WRITING
    builder.write_doc_serialized(docname, doctree)
    builder.write_doc(docname, doctree)



"""Update annotations info of living objects using type_comments."""

from __future__ import annotations

from typing import TYPE_CHECKING

from sphinx.ext.autodoc._dynamic._type_comments import (
    _update_annotations_using_type_comments,
    not_suppressed,  # NoQA: F401
    signature_from_ast,  # NoQA: F401
)

if TYPE_CHECKING:
    from typing import Any

    from sphinx.application import Sphinx


# Retained: legacy class-based
def update_annotations_using_type_comments(
    app: Sphinx, obj: Any, bound_method: bool
) -> None:
    if not app.config.autodoc_use_type_comments:
        return None

    return _update_annotations_using_type_comments(obj, bound_method)



"""Build documentation from a provided source."""

from __future__ import annotations

import argparse
import contextlib
import locale
import multiprocessing
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import sphinx._cli.util.errors
import sphinx.locale
from sphinx import __display_version__
from sphinx._cli.util.colour import disable_colour, terminal_supports_colour
from sphinx.application import Sphinx
from sphinx.locale import __
from sphinx.util._io import TeeStripANSI
from sphinx.util._pathlib import _StrPath
from sphinx.util.docutils import docutils_namespace, patch_docutils
from sphinx.util.osutil import ensuredir

if TYPE_CHECKING:
    from collections.abc import Collection, Sequence
    from typing import Any, TextIO

    from sphinx.extension import Extension


def handle_exception(
    app: Sphinx | None,
    args: Any,
    exception: BaseException,
    stderr: TextIO = sys.stderr,
) -> None:
    if app is not None:
        message_log: Sequence[str] = app.messagelog
        extensions: Collection[Extension] = app.extensions.values()
    else:
        message_log = extensions = ()
    return sphinx._cli.util.errors.handle_exception(
        exception,
        stderr=stderr,
        use_pdb=args.pdb,
        print_traceback=args.verbosity or args.traceback,
        message_log=message_log,
        extensions=extensions,
    )


def jobs_argument(value: str) -> int:
    """Parse the ``--jobs`` flag.

    Return the number of CPUs if 'auto' is used,
    otherwise ensure *value* is a positive integer.
    """
    if value == 'auto':
        return multiprocessing.cpu_count()
    else:
        jobs = int(value)
        if jobs <= 0:
            raise argparse.ArgumentTypeError(
                __('job number should be a positive number')
            )
        else:
            return jobs


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage='%(prog)s [OPTIONS] SOURCEDIR OUTPUTDIR [FILENAMES...]',
        epilog=__('For more information, visit <https://www.sphinx-doc.org/>.'),
        description=__("""
Generate documentation from source files.

sphinx-build generates documentation from the files in SOURCEDIR and places it
in OUTPUTDIR. It looks for 'conf.py' in SOURCEDIR for the configuration
settings. The 'sphinx-quickstart' tool may be used to generate template files,
including 'conf.py'

sphinx-build can create documentation in different formats. A format is
selected by specifying the builder name on the command line; it defaults to
HTML. Builders can also perform other tasks related to documentation
processing.

By default, everything that is outdated is built. Output only for selected
files can be built by specifying individual filenames.
"""),
    )

    parser.add_argument(
        '--version',
        action='version',
        dest='show_version',
        version=f'%(prog)s {__display_version__}',
    )

    parser.add_argument(
        'sourcedir', metavar='SOURCE_DIR', help=__('path to documentation source files')
    )
    parser.add_argument(
        'outputdir', metavar='OUTPUT_DIR', help=__('path to output directory')
    )
    parser.add_argument(
        'filenames',
        nargs='*',
        help=__(
            '(optional) a list of specific files to rebuild. '
            'Ignored if --write-all is specified'
        ),
    )

    group = parser.add_argument_group(__('general options'))
    group.add_argument(
        '--builder',
        '-b',
        metavar='BUILDER',
        dest='builder',
        default='html',
        help=__("builder to use (default: 'html')"),
    )
    group.add_argument(
        '--jobs',
        '-j',
        metavar='N',
        default=1,
        type=jobs_argument,
        dest='jobs',
        help=__(
            'run in parallel with N processes, when possible. '
            "'auto' uses the number of CPU cores"
        ),
    )
    group.add_argument(
        '--write-all',
        '-a',
        action='store_true',
        dest='force_all',
        help=__('write all files (default: only write new and changed files)'),
    )
    group.add_argument(
        '--fresh-env',
        '-E',
        action='store_true',
        dest='freshenv',
        help=__("don't use a saved environment, always read all files"),
    )

    group = parser.add_argument_group(__('path options'))
    group.add_argument(
        '--doctree-dir',
        '-d',
        metavar='PATH',
        dest='doctreedir',
        help=__(
            'directory for doctree and environment files '
            '(default: OUTPUT_DIR/.doctrees)'
        ),
    )
    group.add_argument(
        '--conf-dir',
        '-c',
        metavar='PATH',
        dest='confdir',
        help=__('directory for the configuration file (conf.py) (default: SOURCE_DIR)'),
    )

    group = parser.add_argument_group('build configuration options')
    group.add_argument(
        '--isolated',
        '-C',
        action='store_true',
        dest='noconfig',
        help=__('use no configuration file, only use settings from -D options'),
    )
    group.add_argument(
        '--define',
        '-D',
        metavar='setting=value',
        action='append',
        dest='define',
        default=[],
        help=__('override a setting in configuration file'),
    )
    group.add_argument(
        '--html-define',
        '-A',
        metavar='name=value',
        action='append',
        dest='htmldefine',
        default=[],
        help=__('pass a value into HTML templates'),
    )
    group.add_argument(
        '--tag',
        '-t',
        metavar='TAG',
        action='append',
        dest='tags',
        default=[],
        help=__('define tag: include "only" blocks with TAG'),
    )
    group.add_argument(
        '--nitpicky',
        '-n',
        action='store_true',
        dest='nitpicky',
        help=__('nitpicky mode: warn about all missing references'),
    )

    group = parser.add_argument_group(__('console output options'))
    group.add_argument(
        '--verbose',
        '-v',
        action='count',
        dest='verbosity',
        default=0,
        help=__('increase verbosity (can be repeated)'),
    )
    group.add_argument(
        '--quiet',
        '-q',
        action='store_true',
        dest='quiet',
        help=__('no output on stdout, just warnings on stderr'),
    )
    group.add_argument(
        '--silent',
        '-Q',
        action='store_true',
        dest='really_quiet',
        help=__('no output at all, not even warnings'),
    )
    group.add_argument(
        '--color',
        action='store_const',
        dest='color',
        const='yes',
        default='auto',
        help=__('do emit colored output (default: auto-detect)'),
    )
    group.add_argument(
        '--no-color',
        '-N',
        action='store_const',
        dest='color',
        const='no',
        help=__('do not emit colored output (default: auto-detect)'),
    )

    group = parser.add_argument_group(__('warning control options'))
    group.add_argument(
        '--warning-file',
        '-w',
        metavar='FILE',
        dest='warnfile',
        help=__('write warnings (and errors) to given file'),
    )
    group.add_argument(
        '--fail-on-warning',
        '-W',
        action='store_true',
        dest='warningiserror',
        help=__('turn warnings into errors'),
    )
    group.add_argument('--keep-going', action='store_true', help=argparse.SUPPRESS)
    group.add_argument(
        '--show-traceback',
        '-T',
        action='store_true',
        dest='traceback',
        help=__('show full traceback on exception'),
    )
    group.add_argument(
        '--pdb', '-P', action='store_true', dest='pdb', help=__('run Pdb on exception')
    )
    group.add_argument(
        '--exception-on-warning',
        action='store_true',
        dest='exception_on_warning',
        help=__('raise an exception on warnings'),
    )

    if parser.prog == '__main__.py':
        parser.prog = 'sphinx-build'

    return parser


def make_main(argv: Sequence[str]) -> int:
    """Sphinx build "make mode" entry."""
    from sphinx.cmd import make_mode

    return make_mode.run_make_mode(argv[1:])


def _parse_arguments(
    parser: argparse.ArgumentParser, argv: Sequence[str]
) -> argparse.Namespace:
    args = parser.parse_args(argv)
    return args


def _parse_confdir(noconfig: bool, confdir: str, sourcedir: str) -> str | None:
    if noconfig:
        return None
    elif not confdir:
        return sourcedir
    return confdir


def _parse_doctreedir(doctreedir: str, outputdir: str) -> _StrPath:
    if doctreedir:
        return _StrPath(doctreedir)
    return _StrPath(outputdir, '.doctrees')


def _validate_filenames(
    parser: argparse.ArgumentParser,
    force_all: bool,
    filenames: list[str],
) -> None:
    if force_all and filenames:
        parser.error(__('cannot combine -a option and filenames'))


def _validate_colour_support(colour: str) -> None:
    if colour == 'no' or (colour == 'auto' and not terminal_supports_colour()):
        disable_colour()


def _parse_logging(
    parser: argparse.ArgumentParser,
    quiet: bool,
    really_quiet: bool,
    warnfile: str | None,
) -> tuple[TextIO | None, TextIO | None, TextIO, TextIO | None]:
    status: TextIO | None = sys.stdout
    warning: TextIO | None = sys.stderr
    error = sys.stderr

    if quiet:
        status = None

    if really_quiet:
        status = warning = None

    warnfp = None
    if warning and warnfile:
        try:
            warn_file = Path(warnfile).resolve()
            ensuredir(warn_file.parent)
            # the caller is responsible for closing this file descriptor
            warnfp = open(warn_file, 'w', encoding='utf-8')  # NoQA: SIM115
        except Exception as exc:
            parser.error(__("cannot open warning file '%s': %s") % (warn_file, exc))
        warning = TeeStripANSI(warning, warnfp)  # type: ignore[assignment]
        error = warning

    return status, warning, error, warnfp  # ty: ignore[invalid-return-type]


def _parse_confoverrides(
    parser: argparse.ArgumentParser,
    define: list[str],
    htmldefine: list[str],
    nitpicky: bool,
) -> dict[str, Any]:
    confoverrides: dict[str, Any] = {}
    val: Any
    for val in define:
        try:
            key, _, val = val.partition('=')
        except ValueError:
            parser.error(__('-D option argument must be in the form name=value'))
        confoverrides[key] = val

    for val in htmldefine:
        try:
            key, _, val = val.partition('=')
        except ValueError:
            parser.error(__('-A option argument must be in the form name=value'))
        with contextlib.suppress(ValueError):
            val = int(val)

        confoverrides[f'html_context.{key}'] = val

    if nitpicky:
        confoverrides['nitpicky'] = True

    return confoverrides


def build_main(argv: Sequence[str]) -> int:
    """Sphinx build "main" command-line entry."""
    parser = get_parser()
    args = _parse_arguments(parser, argv)
    args.confdir = _parse_confdir(args.noconfig, args.confdir, args.sourcedir)
    args.doctreedir = _parse_doctreedir(args.doctreedir, args.outputdir)
    _validate_filenames(parser, args.force_all, args.filenames)
    _validate_colour_support(args.color)
    args.status, args.warning, args.error, warnfp = _parse_logging(
        parser, args.quiet, args.really_quiet, args.warnfile
    )
    args.confoverrides = _parse_confoverrides(
        parser, args.define, args.htmldefine, args.nitpicky
    )

    app = None
    try:
        confdir = args.confdir or args.sourcedir
        with patch_docutils(confdir), docutils_namespace():
            app = Sphinx(
                srcdir=args.sourcedir,
                confdir=args.confdir,
                outdir=args.outputdir,
                doctreedir=args.doctreedir,
                buildername=args.builder,
                confoverrides=args.confoverrides,
                status=args.status,
                warning=args.warning,
                freshenv=args.freshenv,
                warningiserror=args.warningiserror,
                tags=args.tags,
                verbosity=args.verbosity,
                parallel=args.jobs,
                keep_going=False,
                pdb=args.pdb,
                exception_on_warning=args.exception_on_warning,
            )
            app.build(args.force_all, args.filenames)
            return app.statuscode
    except (Exception, KeyboardInterrupt) as exc:
        if app is not None:
            message_log: Sequence[str] = app.messagelog
            extensions: Collection[Extension] = app.extensions.values()
        else:
            message_log = extensions = ()
        sphinx._cli.util.errors.handle_exception(
            exc,
            stderr=args.error,
            use_pdb=args.pdb,
            print_traceback=args.verbosity or args.traceback,
            message_log=message_log,
            extensions=extensions,
        )
        return 2
    finally:
        if warnfp is not None:
            # close the file descriptor for the warnings file opened by Sphinx
            warnfp.close()


def _bug_report_info() -> int:
    from platform import platform, python_implementation

    import docutils
    import jinja2
    import pygments

    print('Please paste all output below into the bug report template\n\n')
    print('```text')
    print(f'Platform:              {sys.platform}; ({platform()})')
    print(f'Python version:        {sys.version})')
    print(f'Python implementation: {python_implementation()}')
    print(f'Sphinx version:        {sphinx.__display_version__}')
    print(f'Docutils version:      {docutils.__version__}')
    print(f'Jinja2 version:        {jinja2.__version__}')
    print(f'Pygments version:      {pygments.__version__}')
    print('```')
    return 0


def main(argv: Sequence[str] = (), /) -> int:
    locale.setlocale(locale.LC_ALL, '')
    sphinx.locale.init_console()

    if not argv:
        argv = sys.argv[1:]

    # Allow calling as 'python -m sphinx build …'
    if argv[:1] == ['build']:
        argv = argv[1:]

    if argv[:1] == ['--bug-report']:
        return _bug_report_info()
    if argv[:1] == ['-M']:
        from sphinx.cmd import make_mode

        return make_mode.run_make_mode(argv[1:])
    else:
        return build_main(argv)


if __name__ == '__main__':
    raise SystemExit(main(sys.argv[1:]))



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

from sphinx.locale import __
from sphinx.util import logging

if TYPE_CHECKING:
    from collections.abc import Sequence, Set
    from pathlib import Path
    from typing import Final, Self

    from sphinx.config import Config

LOGGER: Final[logging.SphinxLoggerAdapter] = logging.getLogger('sphinx.ext.apidoc')


def _remove_old_files(
    written_files: Sequence[Path], dest_dir: Path, suffix: str
) -> None:
    files_to_keep = frozenset(written_files)
    for existing in dest_dir.rglob(f'*.{suffix}'):
        if existing not in files_to_keep:
            try:
                existing.unlink()
            except OSError as exc:
                LOGGER.warning(
                    __('Failed to remove %s: %s'),
                    existing,
                    exc.strerror,
                    type='autodoc',
                )


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class ApidocOptions:
    """Options for apidoc."""

    dest_dir: Path
    module_path: Path

    exclude_pattern: Sequence[str] = ()
    max_depth: int = 4
    follow_links: bool = False
    separate_modules: bool = False
    include_private: bool = False
    toc_file: str = 'modules'
    no_headings: bool = False
    module_first: bool = False
    implicit_namespaces: bool = False
    automodule_options: Set[str] = dataclasses.field(default_factory=set)
    suffix: str = 'rst'

    remove_old: bool = True

    quiet: bool = False
    dry_run: bool = False
    force: bool = True

    # --full only
    full: bool = False
    append_syspath: bool = False
    header: str = ''
    author: str | None = None
    version: str | None = None
    release: str | None = None
    extensions: Sequence[str] | None = None
    template_dir: str | None = None


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class ApidocDefaults:
    """Default values for apidoc options."""

    exclude_patterns: list[str]
    automodule_options: frozenset[str]
    max_depth: int
    follow_links: bool
    separate_modules: bool
    include_private: bool
    no_headings: bool
    module_first: bool
    implicit_namespaces: bool

    @classmethod
    def from_config(cls, config: Config, /) -> Self:
        """Collect the default values for apidoc options."""
        return cls(
            exclude_patterns=config.apidoc_exclude_patterns,
            automodule_options=frozenset(config.apidoc_automodule_options),
            max_depth=config.apidoc_max_depth,
            follow_links=config.apidoc_follow_links,
            separate_modules=config.apidoc_separate_modules,
            include_private=config.apidoc_include_private,
            no_headings=config.apidoc_no_headings,
            module_first=config.apidoc_module_first,
            implicit_namespaces=config.apidoc_implicit_namespaces,
        )



"""Utilities for Sphinx extensions."""

from __future__ import annotations

from typing import TYPE_CHECKING

from packaging.version import InvalidVersion, Version

from sphinx.errors import VersionRequirementError
from sphinx.locale import __
from sphinx.util import logging

if TYPE_CHECKING:
    from typing import Any

    from sphinx.application import Sphinx
    from sphinx.config import Config
    from sphinx.util.typing import ExtensionMetadata

logger = logging.getLogger(__name__)


class Extension:
    def __init__(self, name: str, module: Any, **kwargs: Any) -> None:
        self.name = name
        self.module = module
        self.metadata: ExtensionMetadata = kwargs  # type: ignore[assignment]
        self.version = kwargs.pop('version', 'unknown version')

        # The extension supports parallel read or not.  The default value
        # is ``None``.  It means the extension does not tell the status.
        # It will be warned on parallel reading.
        self.parallel_read_safe = kwargs.pop('parallel_read_safe', None)

        # The extension supports parallel write or not.  The default value
        # is ``True``.  Sphinx writes parallelly documents even if
        # the extension does not tell its status.
        self.parallel_write_safe = kwargs.pop('parallel_write_safe', True)


def verify_needs_extensions(app: Sphinx, config: Config) -> None:
    """Check that extensions mentioned in :confval:`needs_extensions` satisfy the version
    requirement, and warn if an extension is not loaded.

    Warns if an extension in :confval:`needs_extension` is not loaded.

    :raises VersionRequirementError: if the version of an extension in
    :confval:`needs_extension` is unknown or older than the required version.
    """
    if config.needs_extensions is None:
        return

    for extname, reqversion in config.needs_extensions.items():
        extension = app.extensions.get(extname)
        if extension is None:
            logger.warning(
                __(
                    'The %s extension is required by needs_extensions settings, '
                    'but it is not loaded.'
                ),
                extname,
            )
            continue

        fulfilled = True
        if extension.version == 'unknown version':
            fulfilled = False
        else:
            try:
                if Version(reqversion) > Version(extension.version):
                    fulfilled = False
            except InvalidVersion:
                if reqversion > extension.version:
                    fulfilled = False

        if not fulfilled:
            raise VersionRequirementError(
                __(
                    'This project needs the extension %s at least in '
                    'version %s and therefore cannot be built with '
                    'the loaded version (%s).'
                )
                % (extname, reqversion, extension.version)
            )


def setup(app: Sphinx) -> ExtensionMetadata:
    app.connect('config-inited', verify_needs_extensions, priority=800)

    return {
        'version': 'builtin',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from docutils.utils import assemble_option_dict

from sphinx.deprecation import RemovedInSphinx11Warning
from sphinx.ext.autodoc._sentinels import ALL, EMPTY, SUPPRESS
from sphinx.locale import __

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping, Set
    from typing import Any, Final, Literal, Self

    from sphinx.ext.autodoc._property_types import _AutodocObjType
    from sphinx.ext.autodoc._sentinels import ALL_T, EMPTY_T, SUPPRESS_T
    from sphinx.util.typing import OptionSpec


# common option names for autodoc directives
AUTODOC_DEFAULT_OPTIONS = (
    'members',
    'undoc-members',
    'no-index',
    'no-index-entry',
    'inherited-members',
    'show-inheritance',
    'private-members',
    'special-members',
    'ignore-module-all',
    'exclude-members',
    'member-order',
    'imported-members',
    'class-doc-from',
    'no-value',
)

AUTODOC_EXTENDABLE_OPTIONS = frozenset({
    'members',
    'private-members',
    'special-members',
    'exclude-members',
})


class _AutoDocumenterOptions:
    # TODO: make immutable.

    no_index: Literal[True] | None = None
    no_index_entry: Literal[True] | None = None
    _tab_width: int = 8

    # module-like options
    members: ALL_T | list[str] | None = None
    undoc_members: Literal[True] | None = None
    inherited_members: Set[str] | None = None
    show_inheritance: Literal[True] | None = None
    synopsis: str | None = None
    platform: str | None = None
    deprecated: Literal[True] | None = None
    member_order: Literal['alphabetical', 'bysource', 'groupwise'] | None = None
    exclude_members: EMPTY_T | set[str] | None = None
    private_members: ALL_T | list[str] | None = None
    special_members: ALL_T | list[str] | None = None
    imported_members: Literal[True] | None = None
    ignore_module_all: Literal[True] | None = None
    no_value: Literal[True] | None = None

    # class-like options (class, exception)
    class_doc_from: Literal['both', 'class', 'init'] | None = None

    # assignment-like (data, attribute)
    annotation: SUPPRESS_T | str | None = None

    noindex: Literal[True] | None = None

    def __init__(self, **kwargs: Any) -> None:
        vars(self).update(kwargs)

    def __repr__(self) -> str:
        args = ', '.join(f'{k}={v!r}' for k, v in vars(self).items())
        return f'_AutoDocumenterOptions({args})'

    def __getattr__(self, name: str) -> object:
        return None  # return None for missing attributes

    def copy(self) -> Self:
        return self.__class__(**vars(self))

    @classmethod
    def from_directive_options(cls, opts: Mapping[str, Any], /) -> Self:
        return cls(**{k.replace('-', '_'): v for k, v in opts.items() if v is not None})

    # Mapping interface:

    def __getitem__(self, item: str) -> Any:
        warnings.warn(
            'The mapping interface for autodoc options objects is deprecated, '
            'and will be removed in Sphinx 11. Use attribute access instead.',
            RemovedInSphinx11Warning,
            stacklevel=2,
        )
        try:
            return getattr(self, item)
        except AttributeError:
            raise KeyError(item) from None

    def __setitem__(self, key: str, value: Any) -> None:
        msg = f'{self.__class__.__name__!r} object does not support indexed assignment'
        raise TypeError(msg)

    def __delitem__(self, key: str) -> None:
        msg = f'{self.__class__.__name__!r} object does not support indexed deletion'
        raise TypeError(msg)

    def __contains__(self, item: str) -> bool:
        warnings.warn(
            'The mapping interface for autodoc options objects is deprecated, '
            'and will be removed in Sphinx 11. Use attribute access instead.',
            RemovedInSphinx11Warning,
            stacklevel=2,
        )
        return hasattr(self, item)

    def __keys(self) -> list[str]:
        return [key for key in dir(self) if not key.startswith('_')]

    def __iter__(self) -> Iterator[str]:
        warnings.warn(
            'The mapping interface for autodoc options objects is deprecated, '
            'and will be removed in Sphinx 11. Use attribute access instead.',
            RemovedInSphinx11Warning,
            stacklevel=2,
        )
        yield from self.__keys()

    def __len__(self) -> int:
        warnings.warn(
            'The mapping interface for autodoc options objects is deprecated, '
            'and will be removed in Sphinx 11. Use attribute access instead.',
            RemovedInSphinx11Warning,
            stacklevel=2,
        )
        return len(self.__keys())

    def keys(self) -> Iterable[str]:
        warnings.warn(
            'The mapping interface for autodoc options objects is deprecated, '
            'and will be removed in Sphinx 11. Use attribute access instead.',
            RemovedInSphinx11Warning,
            stacklevel=2,
        )
        yield from self.__keys()

    def items(self) -> Iterable[tuple[str, Any]]:
        warnings.warn(
            'The mapping interface for autodoc options objects is deprecated, '
            'and will be removed in Sphinx 11. Use attribute access instead.',
            RemovedInSphinx11Warning,
            stacklevel=2,
        )
        for key in self.__keys():
            yield key, getattr(self, key)

    def values(self) -> Iterable[Any]:
        warnings.warn(
            'The mapping interface for autodoc options objects is deprecated, '
            'and will be removed in Sphinx 11. Use attribute access instead.',
            RemovedInSphinx11Warning,
            stacklevel=2,
        )
        for key in self.__keys():
            yield getattr(self, key)

    def get(self, key: str, default: Any | None = None) -> Any | None:
        warnings.warn(
            'The mapping interface for autodoc options objects is deprecated, '
            'and will be removed in Sphinx 11. Use attribute access instead.',
            RemovedInSphinx11Warning,
            stacklevel=2,
        )
        try:
            return getattr(self, key)
        except AttributeError:
            return default


def identity(x: Any) -> Any:
    return x


def members_option(arg: str | None) -> ALL_T | list[str] | None:
    """Used to convert the :members: option to auto directives."""
    if arg is None or arg is True:
        return ALL
    if arg is False:
        return None
    return [stripped for x in arg.split(',') if (stripped := x.strip())]


def exclude_members_option(arg: str | None) -> EMPTY_T | set[str]:
    """Used to convert the :exclude-members: option."""
    if arg is None or arg is True:
        return EMPTY
    return {stripped for x in arg.split(',') if (stripped := x.strip())}


def inherited_members_option(arg: str | None) -> set[str]:
    """Used to convert the :inherited-members: option to auto directives."""
    if arg is None or arg is True:
        return {'object'}
    if arg:
        return {x.strip() for x in arg.split(',')}
    return set()


def member_order_option(
    arg: str | None,
) -> Literal['alphabetical', 'bysource', 'groupwise'] | None:
    """Used to convert the :member-order: option to auto directives."""
    if arg is None or arg is True:
        return None
    if arg in {'alphabetical', 'bysource', 'groupwise'}:
        return arg  # type: ignore[return-value]
    raise ValueError(__('invalid value for member-order option: %s') % arg)


def class_doc_from_option(arg: str | None) -> Literal['both', 'class', 'init']:
    """Used to convert the :class-doc-from: option to autoclass directives."""
    if arg in {'both', 'class', 'init'}:
        return arg  # type: ignore[return-value]
    raise ValueError(__('invalid value for class-doc-from option: %s') % arg)


def annotation_option(arg: str | None) -> SUPPRESS_T | str | Literal[False]:
    if arg is None or arg is True:
        # suppress showing the representation of the object
        return SUPPRESS
    return arg


def bool_option(arg: str | None) -> bool:
    """Used to convert flag options to auto directives.  (Instead of
    directives.flag(), which returns None).
    """
    return True


def merge_members_option(options: dict[str, Any]) -> None:
    """Merge :private-members: and :special-members: options to the
    :members: option.
    """
    if options.get('members') is ALL:
        # merging is not needed when members: ALL
        return

    members = options.setdefault('members', [])
    for key in ('private-members', 'special-members'):
        other_members = options.get(key)
        if other_members is not None and other_members is not ALL:
            for member in other_members:
                if member not in members:
                    members.append(member)


_OPTION_SPEC_COMMON: Final[OptionSpec] = {
    'no-index': bool_option,
    'no-index-entry': bool_option,
}
_OPTION_SPEC_HAS_MEMBERS: Final[OptionSpec] = _OPTION_SPEC_COMMON | {
    'members': members_option,
    'exclude-members': exclude_members_option,
    'undoc-members': bool_option,
    'private-members': members_option,
    'special-members': members_option,
    'member-order': member_order_option,
}
_OPTION_SPEC_MODULE_SPECIFIC: Final[OptionSpec] = {
    'ignore-module-all': bool_option,
    'imported-members': bool_option,
    'deprecated': bool_option,
    'platform': identity,
    'synopsis': identity,
}
_OPTION_SPEC_CLASS_SPECIFIC: Final[OptionSpec] = {
    'class-doc-from': class_doc_from_option,
    'show-inheritance': bool_option,
    'inherited-members': inherited_members_option,
}
_OPTION_SPEC_ASSIGNMENT: Final[OptionSpec] = _OPTION_SPEC_COMMON | {
    'annotation': annotation_option,
    'no-value': bool_option,
}
_OPTION_SPEC_DEPRECATED: Final[OptionSpec] = {
    'noindex': bool_option,
}
_OPTION_SPEC_FUNCTION_DEF: Final = _OPTION_SPEC_COMMON | _OPTION_SPEC_DEPRECATED
_OPTION_SPECS: Final[Mapping[_AutodocObjType, OptionSpec]] = {
    'module': _OPTION_SPEC_HAS_MEMBERS
    | _OPTION_SPEC_MODULE_SPECIFIC
    | {'show-inheritance': bool_option}  # special case
    | {'inherited-members': inherited_members_option}  # special case
    | {'no-value': bool_option}  # special case
    | _OPTION_SPEC_DEPRECATED,
    'class': _OPTION_SPEC_HAS_MEMBERS
    | _OPTION_SPEC_CLASS_SPECIFIC
    | _OPTION_SPEC_DEPRECATED,
    'exception': _OPTION_SPEC_HAS_MEMBERS
    | _OPTION_SPEC_CLASS_SPECIFIC
    | _OPTION_SPEC_DEPRECATED,
    'function': _OPTION_SPEC_FUNCTION_DEF,
    'decorator': _OPTION_SPEC_FUNCTION_DEF,
    'method': _OPTION_SPEC_FUNCTION_DEF,
    'property': _OPTION_SPEC_FUNCTION_DEF,
    'attribute': _OPTION_SPEC_ASSIGNMENT | _OPTION_SPEC_DEPRECATED,
    'data': _OPTION_SPEC_ASSIGNMENT | _OPTION_SPEC_DEPRECATED,
    'type': _OPTION_SPEC_ASSIGNMENT,
}


def _process_documenter_options(
    *,
    obj_type: _AutodocObjType,
    default_options: Mapping[str, str | bool],
    options: dict[str, str | None],
) -> _AutoDocumenterOptions:
    """Recognize options of object type from user input."""
    option_spec = _OPTION_SPECS[obj_type]
    for name in AUTODOC_DEFAULT_OPTIONS:
        if name not in option_spec:
            continue

        negated = options.pop(f'no-{name}', True) is None
        if name in default_options and not negated:
            if name in options and isinstance(default_options[name], str):
                # take value from options if present or extend it
                # with autodoc_default_options if necessary
                if name in AUTODOC_EXTENDABLE_OPTIONS:
                    opt_value = options[name]
                    if opt_value is not None and opt_value.startswith('+'):
                        options[name] = f'{default_options[name]},{opt_value[1:]}'
            else:
                options[name] = default_options[name]  # type: ignore[assignment]
        elif (opt_value := options.get(name)) is not None:
            # remove '+' from option argument if there's nothing to merge it with
            options[name] = opt_value.removeprefix('+')

    opts = assemble_option_dict(options.items(), option_spec)  # type: ignore[arg-type]
    return _AutoDocumenterOptions.from_directive_options(opts)



"""A Base class for additional parsers."""

from __future__ import annotations

from typing import TYPE_CHECKING

import docutils.parsers
import docutils.parsers.rst
from docutils.parsers.rst import states
from docutils.statemachine import StringList
from docutils.transforms.universal import SmartQuotes

from sphinx.deprecation import _deprecation_warning
from sphinx.util.rst import _append_epilogue, _prepend_prologue

if TYPE_CHECKING:
    from docutils import nodes
    from docutils.transforms import Transform

    from sphinx.application import Sphinx
    from sphinx.config import Config
    from sphinx.environment import BuildEnvironment
    from sphinx.util.typing import ExtensionMetadata


class Parser(docutils.parsers.Parser):
    """A base class for source parsers.

    Additional parsers should inherit from this class instead of
    ``docutils.parsers.Parser``.
    This class provides access to core Sphinx objects; *config* and *env*.
    """

    _config: Config
    _env: BuildEnvironment

    @property
    def config(self) -> Config:
        """The config object."""
        cls_module = self.__class__.__module__
        cls_name = self.__class__.__qualname__
        _deprecation_warning(cls_module, f'{cls_name}.config', remove=(10, 0))
        return self._config

    @property
    def env(self) -> BuildEnvironment:
        """The environment object."""
        cls_module = self.__class__.__module__
        cls_name = self.__class__.__qualname__
        _deprecation_warning(cls_module, f'{cls_name}.env', remove=(10, 0))
        return self._env

    def set_application(self, app: Sphinx) -> None:
        """set_application will be called from Sphinx to set app and other instance variables

        :param sphinx.application.Sphinx app: Sphinx application object
        """
        cls_module = self.__class__.__module__
        cls_name = self.__class__.__qualname__
        _deprecation_warning(cls_module, f'{cls_name}.set_application', remove=(10, 0))
        self._config = app.config
        self._env = app.env


class RSTParser(docutils.parsers.rst.Parser, Parser):
    """A reST parser for Sphinx."""

    def get_transforms(self) -> list[type[Transform]]:
        """Sphinx's reST parser replaces a transform class for smart-quotes by its own

        refs: sphinx.io.SphinxStandaloneReader
        """
        transforms = super(RSTParser, RSTParser()).get_transforms()
        transforms.remove(SmartQuotes)
        return transforms

    def parse(self, inputstring: str | StringList, document: nodes.document) -> None:
        """Parse text and generate a document tree."""
        self.setup_parse(inputstring, document)  # type: ignore[arg-type]
        self.statemachine = states.RSTStateMachine(
            state_classes=self.state_classes,
            initial_state=self.initial_state,
            debug=document.reporter.debug_flag,
        )

        # preprocess inputstring
        if isinstance(inputstring, str):
            lines = docutils.statemachine.string2lines(
                inputstring,
                tab_width=document.settings.tab_width,
                convert_whitespace=True,
            )

            inputlines = StringList(lines, document.current_source)
        else:
            inputlines = inputstring

        self.decorate(inputlines)
        self.statemachine.run(inputlines, document, inliner=self.inliner)
        self.finish_parse()

    def decorate(self, content: StringList) -> None:
        """Preprocess reStructuredText content before parsing."""
        _prepend_prologue(content, self._config.rst_prolog)
        _append_epilogue(content, self._config.rst_epilog)


def setup(app: Sphinx) -> ExtensionMetadata:
    app.add_source_parser(RSTParser)

    return {
        'version': 'builtin',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



from __future__ import annotations

from sphinx.ext.autodoc._dynamic._mock import (
    MockFinder,  # NoQA: F401
    MockLoader,  # NoQA: F401
    _make_subclass,  # NoQA: F401
    _MockModule,  # NoQA: F401
    _MockObject,  # NoQA: F401
    ismock,  # NoQA: F401
    ismockmodule,  # NoQA: F401
    mock,  # NoQA: F401
    undecorate,  # NoQA: F401
)



"""Manual pages builder."""

from __future__ import annotations

from typing import TYPE_CHECKING

from sphinx import addnodes
from sphinx._cli.util.colour import darkgreen
from sphinx.builders import Builder
from sphinx.locale import __
from sphinx.util import logging
from sphinx.util.display import progress_message
from sphinx.util.docutils import _get_settings
from sphinx.util.nodes import inline_all_toctrees
from sphinx.util.osutil import ensuredir, make_filename_from_project
from sphinx.writers.manpage import (
    ManualPageTranslator,
    ManualPageWriter,
    NestedInlineTransform,
)

if TYPE_CHECKING:
    from collections.abc import Set

    from sphinx.application import Sphinx
    from sphinx.config import Config
    from sphinx.util.typing import ExtensionMetadata

logger = logging.getLogger(__name__)


class ManualPageBuilder(Builder):
    """Builds groff output in manual page format."""

    name = 'man'
    format = 'man'
    epilog = __('The manual pages are in %(outdir)s.')

    default_translator_class = ManualPageTranslator
    supported_image_types = []

    def init(self) -> None:
        if not self.config.man_pages:
            logger.warning(
                __('no "man_pages" config value found; no manual pages will be written')
            )

    def get_outdated_docs(self) -> str | list[str]:
        return 'all manpages'  # for now

    def get_target_uri(self, docname: str, typ: str | None = None) -> str:
        return ''

    @progress_message(__('writing'))
    def write_documents(self, _docnames: Set[str]) -> None:
        docsettings = _get_settings(
            ManualPageWriter, defaults=self.env.settings, read_config_files=True
        )

        for info in self.config.man_pages:
            docname, name, description, authors, section = info
            if docname not in self.env.all_docs:
                logger.warning(
                    __('"man_pages" config value references unknown document %s'),
                    docname,
                )
                continue
            if isinstance(authors, str):
                if authors:
                    authors = [authors]
                else:
                    authors = []

            docsettings.title = name
            docsettings.subtitle = description
            docsettings.authors = authors
            docsettings.section = section

            if self.config.man_make_section_directory:
                dirname = 'man%s' % section
                ensuredir(self.outdir / dirname)
                targetname = f'{dirname}/{name}.{section}'
            else:
                targetname = f'{name}.{section}'

            logger.info('%s { ', darkgreen(targetname))

            tree = self.env.get_doctree(docname)
            docnames: set[str] = set()
            largetree = inline_all_toctrees(
                self, docnames, docname, tree, darkgreen, [docname]
            )
            largetree.settings = docsettings
            logger.info('} ', nonl=True)
            self.env.resolve_references(largetree, docname, self)
            # remove pending_xref nodes
            for pendingnode in largetree.findall(addnodes.pending_xref):
                pendingnode.replace_self(pendingnode.children)

            transform = NestedInlineTransform(largetree)
            transform.apply()
            visitor: ManualPageTranslator = self.create_translator(largetree, self)  # type: ignore[assignment]
            largetree.walkabout(visitor)
            (self.outdir / targetname).write_text(visitor.astext(), encoding='utf-8')

    def finish(self) -> None:
        pass


def default_man_pages(config: Config) -> list[tuple[str, str, str, list[str], int]]:
    """Better default man_pages settings."""
    filename = make_filename_from_project(config.project)
    return [
        (
            config.root_doc,
            filename,
            f'{config.project} {config.release}',
            [config.author],
            1,
        )
    ]


def setup(app: Sphinx) -> ExtensionMetadata:
    app.add_builder(ManualPageBuilder)

    app.add_config_value(
        'man_pages', default_man_pages, '', types=frozenset({list, tuple})
    )
    app.add_config_value('man_show_urls', False, '', types=frozenset({bool}))
    app.add_config_value(
        'man_make_section_directory', False, '', types=frozenset({bool})
    )

    return {
        'version': 'builtin',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }
